@@ -719,15 +719,16 @@
                        (query-to-vec "SELECT * FROM certname_facts WHERE certname = ?")
                        (reduce #(assoc %1 (:name %2) (:value %2)) {}))
             environment (->> environment_id
-                          (query-to-vec "SELECT name FROM environments WHERE id = ?")
-                          first
-                          :name)]
-        (scf-store/add-facts!
-          {:name (str certname)
-           :values facts
-           :timestamp timestamp
-           :environment environment
-           :producer-timestamp nil})))))
+                             (query-to-vec "SELECT name FROM environments WHERE id = ?")
+                             first
+                             :name)]
+        (when-not (empty? facts)
+          (scf-store/add-facts!
+            {:name (str certname)
+             :values facts
+             :timestamp timestamp
+             :environment environment
+             :producer-timestamp nil}))))))
 
 (defn structured-facts []
   ;; -----------
@@ -843,6 +844,45 @@
    "DROP TABLE certname_facts"
    "DROP TABLE certname_facts_metadata"))
 
+(defn structured-facts-deferrable-constraints []
+
+  ;;This migration is switching from a GC background thread to
+  ;;removing fact values as they are orphaned. Before switching to
+  ;;that model (which only cleans up the specific fact values it
+  ;;orphans) we need to clean up any ones that have appeared between
+  ;;the last run and the current run.
+  (sql/delete-rows :fact_values
+                   ["ID NOT IN (SELECT DISTINCT fact_value_id FROM facts)"])
+  (sql/delete-rows :fact_paths
+                   ["ID NOT IN (SELECT path_id FROM fact_values)"])
+
+  (when (scf-utils/postgres?)
+    (sql/do-commands
+
+     "ALTER TABLE fact_values DROP CONSTRAINT fact_values_path_id_fk"
+     (str "ALTER TABLE fact_values ADD CONSTRAINT fact_values_path_id_fk
+           FOREIGN KEY (path_id) REFERENCES fact_paths (id) MATCH SIMPLE
+           ON UPDATE NO ACTION ON DELETE NO ACTION
+           DEFERRABLE")
+
+     "ALTER TABLE facts DROP CONSTRAINT fact_value_id_fk"
+     (str "ALTER TABLE facts ADD CONSTRAINT fact_value_id_fk
+           FOREIGN KEY (fact_value_id) REFERENCES fact_values(id)
+           ON UPDATE NO ACTION  ON DELETE NO ACTION
+           DEFERRABLE")))
+
+  (sql/do-commands
+   "CREATE INDEX fact_value_id_idx ON facts(fact_value_id)"))
+
+(defn switch-value-string-index-to-gin []
+  "This drops the fact_values_string_trgm index so that it can be recreated
+   as a GIN index."
+
+  (when (and (scf-utils/postgres?)
+             (scf-utils/index-exists? "fact_values_string_trgm"))
+    (sql/do-commands
+     "DROP INDEX fact_values_string_trgm")))
+
 ;; The available migrations, as a map from migration version to migration function.
 (def migrations
   {1 initialize-store
@@ -869,7 +909,9 @@
    22 add-environments
    23 add-report-status
    24 add-producer-timestamps
-   25 structured-facts})
+   25 structured-facts
+   26 structured-facts-deferrable-constraints
+   27 switch-value-string-index-to-gin})
 
 (def desired-schema-version (apply max (keys migrations)))
 
@@ -939,8 +981,11 @@
   (when-not (scf-utils/index-exists? "fact_paths_path_trgm")
     (log/info "Creating additional index `fact_paths_path_trgm`")
     (sql/do-commands
-     "CREATE INDEX fact_paths_path_trgm ON fact_paths USING gist (path gist_trgm_ops)"
-     "CREATE INDEX fact_values_string_trgm ON fact_values USING gist (value_string gist_trgm_ops)")))
+     "CREATE INDEX fact_paths_path_trgm ON fact_paths USING gist (path gist_trgm_ops)"))
+  (when-not (scf-utils/index-exists? "fact_values_string_trgm")
+    (log/info "Creating additional index `fact_values_string_trgm`")
+    (sql/do-commands
+     "CREATE INDEX fact_values_string_trgm ON fact_values USING gin (value_string gin_trgm_ops)")))
 
 (defn indexes!
   "Create missing indexes for applicable database platforms."
